import json
import re
from concurrent import futures
from typing import Callable, Iterator

from google.cloud import pubsub_v1, storage

from dapla import AuthClient


class EmptyListError(Exception):
    pass


def _get_list_of_blobs_with_prefix(
    bucket_name: str, folder_prefix: str
) -> Iterator[storage.Blob]:
    """
    Helper function that gets a list of Blob objects in a Google Cloud Storage bucket that has a certain prefix.

    Args:
        bucket_name (str): The name of the Google Cloud Storage bucket to get blobs from.
        folder_prefix (str): The prefix to filter blobs by.

    Returns:
        An list over the `storage.Blob` objects representing the blobs in the specified bucket and with names starting
        with the given prefix.
    """
    google_credentials = AuthClient.fetch_google_credentials()
    storage_client = storage.Client(credentials=google_credentials)
    return list(storage_client.list_blobs(bucket_name, prefix=folder_prefix))


def _generate_pubsub_data(bucket_id: str, object_id: str) -> bytes:
    """
    Helper function that generates the message data to be published to a Google Cloud Pub/Sub topic.

    Args:
        bucket_id (str): The ID of the Google Cloud Storage bucket that contains the object.
        object_id (str): The ID of the object that has been updated.

    Returns:
        bytes: The message data, encoded as a byte string.
    """
    data = {
        "kind": "storage#object",
        "name": f"{bucket_id}/{object_id}",
        "bucket": f"{bucket_id}",
    }
    return json.dumps(data).encode("utf-8")


def _get_callback(
    publish_future: pubsub_v1.publisher.futures.Future,
    blob_name: str,
    timeout: int = 60,
) -> Callable[[pubsub_v1.publisher.futures.Future], None]:
    """
    Helper function that creates a callback function for a Google Cloud Pub/Sub publish future.

    Args:
        publish_future (pubsub_v1.publisher.futures.Future): The future object returned by the publish call.
        blob_name (str): The name of the Google Cloud Storage object that is being published.
        timeout (Optional[int], default=60): The number of seconds to wait for the publish call to succeed before timing out.

    Returns:
        callable: A callback function that handles success or failure of the publish operation.
    """

    def callback(publish_future: pubsub_v1.publisher.futures.Future) -> None:
        try:
            # Wait 60 seconds for the publish call to succeed.
            publish_future.result(timeout=timeout)
        except futures.TimeoutError:
            print(f"Publishing message for {blob_name} timed out.")
            raise

    return callback


def _publish_gcs_objects_to_pubsub(
    project_id: str, bucket_id: str, folder_prefix: str, topic_id: str
):
    """Publishes messages to a Pub/Sub topic for all objects in a Google Cloud Storage bucket with a given prefix.

    Args:
        project_id (str): The ID of the Google Cloud project that the Pub/Sub topic belongs to.
        bucket_id (str): The ID of the Google Cloud Storage bucket.
        folder_prefix (str): The prefix of the folder containing the objects to be published.
        topic_id (str): The ID of the Pub/Sub topic to publish to.
    """
    blob_list = _get_list_of_blobs_with_prefix(bucket_id, folder_prefix)

    if len(blob_list) == 0:
        raise EmptyListError(
            f"There are no files in {bucket_id:} with the given {folder_prefix:}."
        )

    publisher = pubsub_v1.PublisherClient(
        credentials=AuthClient.fetch_google_credentials()
    )
    topic_path = publisher.topic_path(project_id, topic_id)

    publish_futures = []

    for blob in blob_list:
        byte_data = _generate_pubsub_data(bucket_id, blob.name)

        # When you publish a message, the client returns a future.
        publish_future = publisher.publish(
            topic_path,
            data=byte_data,
            payloadFormat="JSON_API_V1",
            bucketId=f"{bucket_id}",
            objectId=f"{blob.name}",
            eventType="DAPLA-REPUBLISH",
        )

        # Non-blocking. Publish failures are handled in the callback function.
        publish_future.add_done_callback(_get_callback(publish_future, blob.name))
        publish_futures.append(publish_future)

    # Wait for all the publish futures to resolve before exiting.
    futures.wait(publish_futures, return_when=futures.ALL_COMPLETED)
    print(f"Messages published to {topic_path}")


def _extract_project_name(project_id):
    """Extracts the project name from a project ID.

    The project ID is expected to be in the format "<project-name>-<unique-id>".
    This function extracts the project name by matching the project ID up to the
    last hyphen, and removing everything after it.

    Args:
        project_id (str): The GCP project ID to extract the name from.

    Returns:
        str: The GCP project name extracted from the project ID.

    Raises:
        ValueError: If the project ID is not in the expected format.
    """
    match = re.match(r"^(.*)-[^-]+$", project_id)
    if match:
        return match.group(1)
    else:
        raise ValueError(
            f"Invalid project ID: {project_id}, The project ID is expected to be in the format "
            f"<project-name>-<unique-id>"
        )


def trigger_source_data_processing(
    project_id: str, source_name: str, folder_prefix: str
):
    """Triggers a source data processing service with every file that has a given prefix.

    Args:
        project_id (str): The ID of Google Cloud project containing the source.
        folder_prefix (str): The folder prefix of the files to be processed.
        source_name (str): The name of source that should process the files.
    """

    project_name = _extract_project_name(project_id)

    bucket_suffix = "-data-kilde"
    bucket_id = f"ssb-{project_name}{bucket_suffix}"
    topic_id = f"update-{source_name}"

    _publish_gcs_objects_to_pubsub(project_id, bucket_id, folder_prefix, topic_id)



from typing import Any, Dict, List, Optional

# import this module to trigger import side-effect and register the pyarrow extension types
import pandas.core.arrays.arrow.extension_types  # noqa: F401
from pandas import DataFrame, read_csv, read_fwf, read_json, read_xml

from .auth import AuthClient
from .files import FileClient


def read_pandas(
    gcs_path: str, file_format: str = "parquet", columns: List[str] = None, **kwargs
) -> DataFrame:
    """
    Convenience method for reading a dataset from a given GCS path and convert it to a Pandas dataframe.
    :param gcs_path: path to the directory or file you want to get the contents of
    :param file_format: the expected file format. All file formats other than "parquet" are delegated to Pandas'
        methods like read_json, read_csv, etc.
    :param columns: If not None, only these columns will be read
    :return: a pandas dataframe

    Other arguments are passed through to the to_pandas method.
    """
    if file_format == "parquet":
        import pyarrow.parquet as pq

        fs = FileClient.get_gcs_file_system()

        # Workaround for https://github.com/apache/arrow/issues/30481
        gcs_path = FileClient._remove_gcs_uri_prefix(gcs_path)

        parquet_ds = pq.ParquetDataset(
            gcs_path,
            filesystem=fs,
            use_legacy_dataset=False,
        )
        return parquet_ds.read_pandas(columns=columns).to_pandas(
            split_blocks=False, self_destruct=True, **kwargs
        )
    elif file_format == "json":
        return read_json(gcs_path, storage_options=get_storage_options(), **kwargs)
    elif file_format == "csv":
        return read_csv(gcs_path, storage_options=get_storage_options(), **kwargs)
    elif file_format == "fwf":
        return read_fwf(gcs_path, storage_options=get_storage_options(), **kwargs)
    elif file_format == "xml":
        return read_xml(gcs_path, storage_options=get_storage_options(), **kwargs)
    else:
        raise ValueError(f"Invalid file format {file_format}")


def write_pandas(
    df: DataFrame, gcs_path: str, file_format: str = "parquet", **kwargs
) -> DataFrame:
    """
    Convenience method for writing a pandas dataframe to a given GCS path.
    :param df: the pandas' dataset
    :param gcs_path: path to the file you want to write to - must have a suffix that corresponds to the file_format
    :param file_format: file format of file you want to write to
    Other arguments are passed through to the pyarrow write_table method.
    """
    import pyarrow.parquet

    import pandas.io.parquet

    pandas.io.parquet.BaseImpl.validate_dataframe(df)

    if file_format == "parquet":
        # Transfom and write pandas dataframe
        from_pandas_kwargs = {"schema": kwargs.pop("schema", None)}
        table = pyarrow.Table.from_pandas(df, preserve_index=True, **from_pandas_kwargs)
        if ".parquet" not in gcs_path:
            raise ValueError("Path must be a parquet file")  # noqa: TC003
        fs = FileClient.get_gcs_file_system()
        with fs.open(gcs_path, mode="wb") as buffer:
            pyarrow.parquet.write_table(
                table,
                buffer,
                flavor={"spark"},
                compression="snappy",
                coerce_timestamps="ms",
                **kwargs,
            )
    elif file_format == "json":
        df.to_json(gcs_path, storage_options=get_storage_options(), **kwargs)
    elif file_format == "csv":
        df.to_csv(gcs_path, storage_options=get_storage_options(), **kwargs)
    elif file_format == "xml":
        df.to_xml(gcs_path, storage_options=get_storage_options(), **kwargs)
    else:
        raise ValueError(f"Invalid file format {file_format}")


def get_storage_options() -> Optional[Dict[str, Any]]:
    """
    Returns the ``storage_options`` that are used in Pandas for specifying extra options for a particular storage
    connection that will be parsed by ``fsspec``. In this case when the URL starts with "gcs://".
    An error will be raised by Pandas if providing this argument with a local path or a file-like buffer.
    See the fsspec and backend storage implementation docs for the set of allowed keys and values
    """
    token = AuthClient.fetch_google_credentials()
    return {"token": token} if token is not None else None



__version__ = "1.7.1"

from .auth import AuthClient
from .backports import details, show
from .collector import CollectorClient
from .converter import ConverterClient
from .doctor import Doctor
from .files import FileClient
from .guardian import GuardianClient
from .jupyterhub import generate_api_token
from .pandas import read_pandas, write_pandas
from .pubsub import trigger_source_data_processing
